from json import loads as json_loads
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, Optional, Tuple, Type, TypeVar

from pydantic import BaseSettings
from pydantic.env_settings import SettingsSourceCallable

from .sources.glob import GlobSource, Loader

toml_loads: Optional[Loader]
try:
    from toml import loads as toml_loads
except ImportError:
    toml_loads = None

cli_settings_source: Optional[Callable[..., SettingsSourceCallable]]
try:
    from .sources.cli import cli_settings_source
except ImportError:
    cli_settings_source = None

SettingsT = TypeVar("SettingsT", bound=BaseSettings)


def autofill(
    name: str,
    *,
    extra_sources: Iterable[SettingsSourceCallable] = tuple(),
    cli_settings: Optional[Dict[str, Any]] = None,
) -> Callable[[Type[SettingsT]], Type[SettingsT]]:
    """Fill in the blanks based on setting files, env vars, etc."""
    if cli_settings is None:
        cli_settings = {}

    def _autofill(base: Type[SettingsT]) -> Type[SettingsT]:
        # Early out if the class already has this decoration
        if hasattr(base, "__autofill__"):
            return base

        class _Wrapper(base):  # type: ignore[valid-type, misc] # pylint: disable=too-few-public-methods
            class Config:  # pylint: disable=missing-docstring,too-few-public-methods
                env_prefix = f"{name}_"

                @classmethod
                def customise_sources(
                    cls,
                    init_settings: SettingsSourceCallable,
                    env_settings: SettingsSourceCallable,
                    file_secret_settings: SettingsSourceCallable,
                ) -> Tuple[SettingsSourceCallable, ...]:
                    """Apply setting sources with custom precedence."""
                    sources = [
                        # First (highest precedence), settings given via the constructor
                        # itself directly within Python. E.g.:
                        #   settings = Settings(debug=True, background=True)
                        init_settings
                    ]
                    if cli_settings_source is not None:
                        assert cli_settings is not None
                        # Second, CLI settings (if you enable this extra). E.g.:
                        #   $ ./appster --debug --background
                        sources.append(cli_settings_source(name, **cli_settings))
                    sources += [
                        # Third, settings from environment variables. E.g.:
                        #   $ APPSTER_DEBUG=y APPSTER_BACKGROUND=y ./appster
                        env_settings,
                        # Fourth, settings from secret files. E.g., files in
                        #   /var/run/database_password
                        #   /var/run/backend_ip_address
                        file_secret_settings,
                    ]
                    if toml_loads is not None:
                        sources += [
                            # Fifth, setting files from the current directory. E.g.:
                            #   ./dev-credentials.appster.toml
                            #   ./z10-relocate-db.appster.toml
                            #   ./z99-disable-ssl.appster.toml
                            # Note that we apply multiple setting files in alphanumeric
                            # order.
                            # Note that TOML files have higher precedence than
                            # JSON files. We may change this in the future. Don't rely
                            # on this behaviour
                            GlobSource(Path("./"), f"*.{name}.toml", toml_loads),
                            GlobSource(Path("./"), f"*.{name}.json", json_loads),
                            # Sixth, setting files from the system's settings
                            # directory. E.g.:
                            #   /etc/appster/base-settings.toml
                            #   /etc/appster/z10-relocate-db.toml
                            #   /etc/appster/z99-disable-ssl.toml
                            # Note that we apply multiple setting files in alphanumeric
                            # order.
                            # Note that TOML files have higher precedence than
                            # JSON files. We may change this in the future. Don't rely
                            # on this behaviour
                            GlobSource(Path(f"/etc/{name}"), "*.toml", toml_loads),
                            GlobSource(Path(f"/etc/{name}"), "*.json", json_loads),
                        ]
                    # Seventh (lowest precedence), you can specify additional setting
                    # sources.
                    sources += extra_sources
                    return tuple(sources)

        # Mark the class so that we know that it is decorated
        _Wrapper.__autofill__ = True
        # Wraps
        _Wrapper.__name__ = base.__name__
        _Wrapper.__doc__ = base.__doc__
        return _Wrapper

    return _autofill



from pathlib import Path
from typing import Any, Callable, Dict, Type, TypeVar

from pydantic import BaseSettings, Extra, Field
from pydantic.env_settings import SettingsSourceCallable

from ..settings import autofill as base_autofill

SettingsT = TypeVar("SettingsT", bound=BaseSettings)


def autofill(name: str) -> Callable[[Type[SettingsT]], Type[SettingsT]]:
    """Fill in the blanks based on setting files, env vars, etc."""
    extra_sources = (_app_computed_settings(name),)
    return base_autofill(name, extra_sources=extra_sources)


def _app_computed_settings(name: str) -> SettingsSourceCallable:
    def _source(_: BaseSettings) -> Dict[str, Any]:
        return {
            "data_directory": Path(f"/var/{name}"),
        }

    return _source


class Settings(BaseSettings):
    """Application base settings."""

    debug: bool = Field(False, description="Enable debug checks.")
    background: bool = Field(
        True,
        disable_name="foreground",
        description="Daemonize this process.",
    )
    data_directory: Path = Field(
        ...,
        description="Where this app stores its data.",
    )

    class Config:  # pylint: disable=too-few-public-methods
        allow_mutation = False
        extra = Extra.forbid



import inspect
from contextlib import AsyncExitStack, suppress
from functools import wraps
from typing import (
    Any,
    AsyncContextManager,
    Callable,
    ContextManager,
    Coroutine,
    Optional,
    Protocol,
    Type,
    TypeVar,
)

from anyio import create_task_group
from anyio.abc import TaskGroup

ReturnT = TypeVar("ReturnT", covariant=True)

Func = Callable[..., Coroutine[Any, Any, ReturnT]]


# Note that we disable D102 for `Protocol`s since it's redundant documentation.
# Similarly, we disable too-few-public-methods since it doens't make sense for
# `Protocol`s. Hopefully, both pydocstyle and pylint will special-case `Protocol`s
# soon enough.


class InjectedFunc(Protocol[ReturnT]):  # pylint: disable=too-few-public-methods
    """`Func` after we apply `inject` to it."""

    async def __call__(self) -> ReturnT:  # noqa: D102
        ...


class Factory(Protocol):  # pylint: disable=too-few-public-methods
    """Given a type, return an instance of said type."""

    async def __call__(self, __annotation: Type[Any]) -> Any:  # noqa: D102
        ...


async def _basic_factory(annotation: Type[Any]) -> Any:
    if issubclass(annotation, TaskGroup):
        return create_task_group()
    raise ValueError


def inject(
    *,
    extra_factory: Optional[Factory] = None,
) -> Callable[[Func[ReturnT]], InjectedFunc[ReturnT]]:
    """Inject instances of the given function's argument types."""

    def _inject(coro: Func[ReturnT]) -> InjectedFunc[ReturnT]:
        @wraps(coro)
        async def _wrapper() -> ReturnT:  # type: ignore[return]
            spec = inspect.getfullargspec(coro)
            args: Any = []
            async with AsyncExitStack() as stack:
                for arg_name in spec.args:
                    try:
                        annotation = spec.annotations[arg_name]
                    except KeyError as exc:
                        raise ValueError(
                            f'Argument "{arg_name} must have a type annotation"'
                        ) from exc

                    arg = await _get_arg(annotation, stack, extra_factory)
                    if arg is None:
                        raise ValueError(
                            f'Argument "{arg_name}" has unknown type '
                            f'annotation "{annotation}"'
                        )

                    # There is a bug in pylint with
                    # isinstance-second-argument-not-valid-type
                    # See: https://github.com/PyCQA/pylint/issues/3507
                    if isinstance(  # pylint: disable=isinstance-second-argument-not-valid-type
                        arg, AsyncContextManager
                    ):
                        arg = await stack.enter_async_context(arg)
                    elif isinstance(  # pylint: disable=isinstance-second-argument-not-valid-type
                        arg, ContextManager
                    ):
                        arg = stack.enter_context(arg)
                    args.append(arg)

                return await coro(*args)

        return _wrapper

    return _inject


async def _get_arg(
    annotation: Type[Any],
    stack: AsyncExitStack,
    extra_factory: Optional[Factory] = None,
) -> Any:
    if issubclass(annotation, AsyncExitStack):
        return stack
    with suppress(ValueError):
        return await _basic_factory(annotation)
    if extra_factory is not None:
        with suppress(ValueError):
            return await extra_factory(annotation)
    return None



from typing import Any


class ExtraError(Exception):
    """Base exception for all extra-related erorrs."""

    def __init__(self, extra_name: str, *args: Any) -> None:
        super().__init__(*args)
        self._extra_name = extra_name


class ExtraImportError(ExtraError, ImportError):
    """The extra depends on a component that isn't available."""

    def __init__(self, extra_name: str) -> None:
        msg = (
            f'The "{extra_name}" extra depends on a component that isn\'t available. '
            f'Did you forget to specify the "{extra_name}" extra during install? '
            f'Try again with, e.g., "poetry install --extras {extra_name}"'
        )
        super().__init__(extra_name, msg)



from __future__ import annotations

import inspect
import logging
from contextlib import ExitStack
from pathlib import Path
from types import TracebackType
from typing import Any, ContextManager, Optional, Type, TypeVar, cast

from anyio import run

from ._inject import Func, inject
from ._settings import Settings, autofill

_LOGGER = logging.getLogger(__name__)


ReturnT = TypeVar("ReturnT")


class App(ContextManager["App"]):
    """Useful defaults for applications."""

    def __init__(self, name: str, settings: Settings) -> None:
        self._name = name
        self._settings = settings
        self._stack = ExitStack()

    @property
    def name(self) -> str:
        """Application name."""
        return self._name

    @property
    def settings(self) -> Settings:
        """Application settings."""
        return self._settings

    @classmethod
    def launch(
        cls,
        func: Func[ReturnT],
        *,
        name: Optional[str] = None,
        settings_class: Optional[Type[Settings]] = None,
    ) -> ReturnT:
        """Create app instance and run the given coroutine function."""
        # Set defaults for optional arguments
        if name is None:
            name = get_app_name(func)
        if settings_class is None:
            settings_class = get_settings_class(func)
        # Automatically fill in missing settings (e.g., from settings files)
        settings_class = autofill(name)(settings_class)
        # Create settings instance
        settings: Settings = settings_class()
        # Create and run app instance
        with cls(name, settings) as app:
            return app.run(func)

    def run(self, func: Func[ReturnT]) -> ReturnT:
        """Execute the coroutine function and return the result."""
        # Inject dependencies (e.g., a task group, settings, stack, etc.)
        injected_func = inject(extra_factory=self._factory)(func)
        # Run until the task completes (or the process receives a signal)
        return run(injected_func)

    async def _factory(self, annotation: Type[Any]) -> Any:
        """Return instance based on the given annotation.

        We use this function for dependency injection.
        """
        if issubclass(annotation, App):
            return self
        if issubclass(annotation, Settings):
            return self._settings
        raise ValueError

    def __enter__(self) -> App:
        try:
            self._stack.__enter__()
        except Exception as error:
            _LOGGER.error("Could not start due to error:", exc_info=error)
            raise
        _LOGGER.info("Started")
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self._stack.__exit__(exc_type, exc_value, traceback)
        # Log exit
        if exc_value is not None:
            # We exit due to a problem
            _LOGGER.error("Stopped due to error:", exc_info=exc_value)
        else:
            _LOGGER.info("Stopped")


def get_settings_class(func: Func[ReturnT]) -> Type[Settings]:
    """Try to get the settings class from the function signature."""
    spec = inspect.getfullargspec(func)
    for arg_name in spec.args:
        try:
            annotation = spec.annotations[arg_name]
        except KeyError:
            continue
        if issubclass(annotation, Settings):
            # Strangely, mypy can't infer that `annotation` has the right
            # type from the `issubclass` call. We explicitly cast as a
            # work-around to this.
            return cast(Type[Settings], annotation)
    # Default to the base settings class
    return Settings


def get_app_name(func: Func[ReturnT]) -> str:
    """Get the name of the running application."""
    # Use the function name itself if said name is descriptive
    if func.__name__ not in ("main",) and not func.__name__.startswith("_"):
        return func.__name__

    # If the function name isn't desciptive, we query the main module
    import __main__ as main  # type: ignore[import]  # pylint: disable=import-outside-toplevel

    # Use the module name if the application runs a module
    #
    # `main` will look something like this:
    #   main.__name__   : __main__
    #   main.__package__: baxter
    #   main.__file__   : /media/system/lib/python3.6/site-packages/baxter/__main__.py
    if isinstance(main.__package__, str):
        return main.__package__

    # Use the file name if the application runs directly
    #
    # `main` will look something like this:
    #   main.__name__   : __main__
    #   main.__package__: None
    #   main.__file__   : /media/system/bin/baxter
    #
    # Or, sometimes, like this:
    #   main.__file__   : _appster.py
    file_name = Path(main.__file__).stem  # Use the stem to avoid file extensions
    # Remove leading and trailing underscores (if any).
    # All in all, a file name like "_appster.py" becomes "appster".
    return file_name.strip("_")



from __future__ import annotations

import json
import sys
from collections import defaultdict
from dataclasses import dataclass
from typing import (
    Any,
    Container,
    DefaultDict,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    get_args,
    get_origin,
)

import click
from pydantic import BaseModel, BaseSettings
from pydantic.env_settings import SettingsSourceCallable
from pydantic.fields import ModelField
from pydantic.types import Json, JsonWrapper
from pydantic.utils import lenient_issubclass


def cli_settings_source(
    name: str, *, delimiter: str = ".", internal_delimiter: str = "__"
) -> SettingsSourceCallable:
    """Return settings source based on this process' CLI options."""
    if not internal_delimiter.isidentifier():  # [1]
        raise ValueError(
            f'The internal delimiter "{internal_delimiter}" '
            "is not a valid identifier."
        )

    def _cli_settings_source(settings: BaseSettings) -> Dict[str, Any]:
        """Return settings from this process' CLI options.

        Limitations:
        * No support for positional arguments

        Examples:
            "--debug"
            {
                "debug": True
            }

            "--no-debug"
            {
                "debug": False
            }

            "--foo.bar /tmp"
            {
                "foo": {
                    "bar": "/tmp"
                }
            }

        """
        result: Dict[str, Any] = {}

        def _set_result(**kwargs: Any) -> None:
            nonlocal result
            # Click provides us with all the user-specified options in `kwargs`.
            # To make the `pydantic.BaseModel` fit into Click's API, we flattened
            # the model hierarchy. Now, we convert `kwargs` back into a nested
            # structure of dicts that we can use as a settings source.
            result = _kwargs_to_settings(kwargs, internal_delimiter)

        params: List[click.Parameter] = list(
            _to_options(settings, delimiter, internal_delimiter)
        )
        command = click.Command(name=name, callback=_set_result, params=params)
        # Per default, click calls `sys.exit` whenever the command is done.
        # We don't want this behaviour, so we disable it with `standalone_mode=False`.
        code: Optional[int] = command.main(standalone_mode=False)
        # If `code` is set, the user asked for help (e.g., via "--help").
        # In this case, we exit the application right away.
        if code is not None:
            sys.exit(code)

        return result

    return _cli_settings_source


def _kwargs_to_settings(
    kwargs: Dict[str, Any], internal_delimiter: str
) -> Dict[str, Any]:
    """Convert flattened kwargs to a nested dict.

    Examples:
        The kwargs:

        `roast_level: 42`
        `preference__cream_and_sugar: True`

        converts to

        `{ "roast_level": 42, "preference": { "cream_and_sugar": True } }`
    """
    result: DefaultDict[str, Any] = defaultdict(dict)
    for full_name, value in kwargs.items():
        # Skip unset values. E.g., for unspecified CLI options.
        if value is None:
            continue
        # Split the full name into parts.
        # E.g.: "preference__cream_and_sugar" into ["preference", "cream_and_sugar"]
        parts = full_name.split(internal_delimiter)
        # Create nested dicts corresponding to each of the parts
        dic = result
        for part in parts[:-1]:  # Skip the last part. It will hold the value itself.
            dic = dic[part]
        # Assign the kwarg value to the innermost dict.
        dic[parts[-1]] = value
    return dict(result)


def _to_options(
    model: Union[BaseModel, Type[BaseModel]],
    delimiter: str,
    internal_delimiter: str,
    *,
    parent_path: Tuple[str, ...] = tuple(),
) -> Iterable[click.Option]:
    """Convert `pydantic.BaseModel` to the equivalent `click.Option`s."""
    # Let's use an example to explain the code below as we go along.
    # A simple model-within-a-model:
    #
    #     class CoffeePreference(BaseModel):
    #         cream_and_sugar: bool
    #
    #     class Coffee(BaseModel):
    #         roast_level: int
    #         preference: CoffeePreference
    #
    # In our example, there are only three fields in total:
    #   1. `roast_level` (from the root model)
    #   2. `preference` (from the root model)
    #   3. `cream_and_sugar` (from the nested model)

    # We go through all the fields of the model.
    for field in model.__fields__.values():
        # Ensure that `field.name` doesn't contain any of the delimiters.
        # Otherwise, the ambiguity makes it impossible to distinguish
        # between nested models and field names.
        _raise_on_delimiter_conflict(field.name, delimiter)
        _raise_on_internal_delimiter_conflict(field.name, internal_delimiter)

        # `field.name` is either:
        #   1. "roast_level"
        #   2. "preference"
        #   3. "cream_and_sugar".
        #
        # In the command line, we prefer kebab case over snake case.
        #
        # `kebab_name` becomes:
        #   1. "roast-level"
        #   2. "preference"
        #   3. "cream-and-sugar".
        kebab_name = field.name.replace("_", "-")

        # Like before, we ensure that `kebab_name` doesn't cause ambiguities.
        _raise_on_delimiter_conflict(kebab_name, delimiter)
        # Note that we don't have to test the internal delimiter since it can't
        # contain the "-" character (see [1]). We assert instead to catch any
        # errors if the check at [1] changes at some point in time.
        assert internal_delimiter not in kebab_name

        # `field.outer_type_` is the type:
        #   1. `int`
        #   2. `CoffeePreference`
        #   3. `bool`
        #
        # Note that sometimes `outer_type_` isn't a type but an instance from the
        # `typing` module. E.g., a `typing.List[int]` instance. Therefore, we use
        # the `lenient_issubclass` utility function instead of `issubclass`. The
        # latter doesn't accept non-types but the former does (hence the leniency).
        if lenient_issubclass(field.outer_type_, BaseModel):
            # Recurse into nested models. E.g., the `CoffeePreference` model
            # from the `preference` field.
            yield from _to_options(
                field.outer_type_,
                delimiter,
                internal_delimiter,
                parent_path=parent_path + (kebab_name,),
            )
            continue
        # CLI-specific extras (settings)
        extras = field.field_info.extra.get("cli", CliExtras())
        # If the field isn't a model itself, we call it a "simple" field.
        # E.g., the `roast_level` and `cream_and_sugar` fields. Note that
        # e.g., `typing.List` and `dict` are also simple fields.
        #
        # We can "clickify" simple fields. That is, convert the field into a
        # `click.Option`.
        #
        # First, we need a so-called parameter declaration for the `click.Option`.
        # A parameter declaration is a set of context-specific names for the option.
        # We generate the parameter declaration from the field.
        param_decls = _ParamDecls.from_field(
            field, kebab_name, delimiter, internal_delimiter, parent_path, extras
        )
        # Then, we "clickify" the field.
        yield _Option.from_field(field, param_decls, extras)


def _raise_on_delimiter_conflict(name: str, delimiter: str) -> None:
    if delimiter in name:
        raise ValueError(
            f'The "{name}" field conflicts with the delimiter "{delimiter}".'
        )


def _raise_on_internal_delimiter_conflict(name: str, internal_delimiter: str) -> None:
    if internal_delimiter in name:
        raise ValueError(
            f'The "{name}" field conflicts with the internal '
            f'delimiter "{internal_delimiter}".'
        )


@dataclass(frozen=True)
class _ParamDecls:
    # The identifier used in the callback.
    # E.g.: "preference__cream_and_sugar".
    identifier: str
    # The name given to the option (as seen on the command line).
    # E.g.: "--preference.cream-and-sugar"
    name: str

    def __post_init__(self) -> None:
        assert self.identifier.isidentifier()

    @classmethod
    def from_field(
        cls,
        field: ModelField,
        kebab_name: str,
        delimiter: str,
        internal_delimiter: str,
        parent_path: Tuple[str, ...],
        extras: CliExtras,
    ) -> _ParamDecls:
        # We use a delimiter (e.g., ".") to separate the names of nested models.
        #
        # `base_option_name` is:
        #   1. "roast-level"
        #   2. "preference.cream-and-sugar".
        base_option_name = delimiter.join(parent_path + (kebab_name,))
        # `click` expects the full option name (with a "--" prefix)
        #
        # `full_option_name` is
        #   1. "--roast-level"
        #   2. "--preference.cream-and-sugar/--preference.no-cream-and-sugar"
        #
        # Note that we automatically add a "disable" option for boolean fields.
        full_option_name = _full_option_name(
            field, kebab_name, base_option_name, delimiter, parent_path, extras
        )
        # `click` requires that the option also has a proper identifier
        # so that we can access the option via keyword argument in the callback.
        # Unfortunately, a name such as "preference.cream-and-sugar" is not
        # a proper identifier due to:
        #  * The delimiter (e.g., ".")
        #  * The "-" character from kebab case
        #
        # Therefore, we replace said chars with:
        #  * The internal delimiter (e.g., "__")
        #  * The "_" character from snake case
        #
        # The resulting string is a proper identifier.
        #
        # In our example, `identifier` is:
        #   1. "roast_level"
        #   2. "preference__cream_and_sugar"
        identifier = base_option_name.replace(delimiter, internal_delimiter).replace(
            "-", "_"
        )
        return cls(identifier, full_option_name)

    def as_tuple(self) -> Tuple[str, str]:
        return (self.identifier, self.name)


def _full_option_name(
    field: ModelField,
    kebab_name: str,
    base_option_name: str,
    delimiter: str,
    parent_path: Tuple[str, ...],
    extras: CliExtras,
) -> str:
    full_option_name = f"--{base_option_name}"
    # Early out of non-boolean fields
    if field.outer_type_ is not bool:
        return full_option_name
    # The user can specify their own "disable" flag
    if extras.disable_flag is not None:
        # Convert into kebab case
        disable_flag = extras.disable_flag.replace("_", "-")
        # Ensure that the disable name is unambiguous
        _raise_on_delimiter_conflict(disable_flag, delimiter)
    # Fall back to a simple "no-" prefix to the "enable" flag
    else:
        disable_flag = f"no-{kebab_name}"
    full_disable_flag = delimiter.join(parent_path + (disable_flag,))
    return full_option_name + f"/--{full_disable_flag}"


class JsonType(click.ParamType):
    """JSON parameter type."""

    name = "json"

    def convert(
        self, value: str, param: Optional[click.Parameter], ctx: Optional[click.Context]
    ) -> Any:
        """Convert raw JSON string to a dict."""
        try:
            return json.loads(value)
        except json.JSONDecodeError as exc:
            self.fail(
                f"'{value}' is invalid JSON. {str(exc)}",
                param,
                ctx,
            )


JSON_TYPE = JsonType()


class _Option(click.Option):
    """A `click.Option` that knows if the user explicitly specified it.

    This allows us to distinguish between the following states:
      * The option has value because it has a default.
      * The option has value because the user explicitly assigned it one
        even if said value coincides with the default.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        if "callback" in kwargs:
            raise ValueError(
                'You can\'t specify the "callback" argument.'
                "We reserve it for internal use."
            )
        super().__init__(
            *args,
            **kwargs,
            # We filter out options that the user did not explicitly specify.
            # Otherwise, we may erroneously give the field's default a higher
            # precedence.
            callback=allow_if_specified,
        )
        self.specified = False

    def handle_parse_result(self, ctx: Any, opts: Any, args: Any) -> Any:
        self.specified = self.name in opts
        return super().handle_parse_result(ctx, opts, args)

    @classmethod
    def from_field(
        cls, field: ModelField, param_decls: _ParamDecls, extras: CliExtras
    ) -> _Option:
        # This function is for simple fields. That is, non-model fields.
        assert not lenient_issubclass(field.outer_type_, BaseModel)
        # Clickify the field's members
        type_ = _clickify_type(field.outer_type_, extras)
        default = _clickify_default(field.default, field.outer_type_, extras)
        show_default = _get_show_default(field.default, field.outer_type_)
        multiple = _get_multiple(field.outer_type_, extras)
        # Return an option constructed from the clickified field members
        return cls(
            param_decls.as_tuple(),
            type=type_,
            default=default,
            show_default=show_default,
            multiple=multiple,
            help=field.field_info.description,
        )


SingleClickParamType = Union[type, click.ParamType]
ClickParamType = Union[SingleClickParamType, Tuple[SingleClickParamType, ...]]


class CliExtras(BaseModel):
    """CLI-specific model settings."""

    # Force click to parse the corresponding option as JSON
    force_json: bool = False
    # Use a custom "disable" flag. Only valid for boolean fields.
    # Defaults to the "enable" name prefixed with "no-". E.g.: "debug" becomes
    # "no-debug".
    # Don't prefix the disable flag with "--".
    disable_flag: Optional[str] = None


def _clickify_type(type_: type, extras: CliExtras) -> ClickParamType:
    # Early out if the user explicitly forces the field type to JSON
    if extras.force_json:
        return JSON_TYPE
    # E.g.: Json, Json[List[str]], etc.
    if lenient_issubclass(type_, (Json, JsonWrapper)):
        # Return `str` since pydantic will parse the JSON in a later step
        return str
    # E.g.: dict, Dict[str, Any], OrderedDict, etc.
    if _is_mapping(type_):
        return JSON_TYPE
    # E.g.: list, FrozenSet[int], Tuple[int, ...], etc.
    if _is_container(type_):
        return _clickify_container_args(type_)
    # E.g.: int, str, float, etc.
    return type_


def _clickify_default(default: Any, type_: type, extras: CliExtras) -> Any:
    # Pydantic uses both `None` and `Ellipsis` to denote "no default value".
    # Click only understands `None`, so we return that.
    if default in (None, Ellipsis):
        return None
    # Early out if the user explicitly forces the field type to JSON
    if extras.force_json:
        return json.dumps(default)
    if _is_mapping(type_):
        return json.dumps(default)
    if _is_container(type_):
        return _clickify_container_default(default)
    return default


def _get_show_default(default: Any, type_: type) -> Union[bool, str]:
    # click's help message for an empty container is "[default: ]". This can
    # confuse the user user. Therefore, we explicitly set the `show_default`
    # to, e.g., "empty list". In turn, click displays this as
    # "[default: (empty list)]". Note that click adds the parentheses.
    if _is_container(type_) and not default:
        name = _type_name(type_)
        return f"empty {name}"
    # For non-containers, we always show the default. We simply return `True` so
    # that click automatically figures out a proper default text.
    return True


def _get_multiple(type_: type, extras: CliExtras) -> bool:
    # Early out if the user explicitly forces the field type to JSON
    if extras.force_json:
        return False
    # Early out for mappings. E.g., dict.
    if _is_mapping(type_):
        return False
    # For containers, we allow multiple arguments. This way, the user
    # can specify an option multiple times and click gathers all values
    # into a single container. E.g.:
    #   $ python app.py --lucky-numbers 2 --lucky-numbers 7
    # becomes
    #   `{ "lucky_numbers": [2, 7] }`.
    if _is_container(type_):
        args = _clickify_container_args(type_)
        # A non-composite type has a single argument.
        # E.g., `List[int]`.
        # A composite type has a tuple of arguments.
        # E.g., `Tuple[str, int, int]`.
        composite = isinstance(args, tuple)
        # We only allow the user to specify multiple values for non-composite types.
        # E.g., for `list`, `Tuple[str, ...]`, `FrozenSet[int]`, etc.
        return not composite
    return False


def _is_mapping(type_: type) -> bool:
    # Early out for standard containers. E.g., dict, OrderedDict
    if lenient_issubclass(type_, Mapping):
        return True
    origin = get_origin(type_)
    # Early out for non-typing objects
    if origin is None:
        return False
    return issubclass(origin, Mapping)


def _is_container(type_: type) -> bool:
    # Early out for `str`. While `str` is technically a container, it's easier to
    # not consider it one in the context of command line interfaces.
    if type_ is str:
        return False
    # Early out for standard containers. E.g.: list, tuple, range
    if lenient_issubclass(type_, Container):
        return True
    origin = get_origin(type_)
    # Early out for non-typing objects
    if origin is None:
        return False
    return issubclass(origin, Container)


def _clickify_container_args(
    type_: type,
) -> ClickParamType:
    assert _is_container(type_)
    args: Tuple[Any, ...] = get_args(type_)
    # Early out for untyped containers such as `tuple`, `List[Any]`, `frozenset`, etc.
    if len(args) == 0:
        # When we don't know the type, we choose `str`. It's tempting to choose `None`
        # but that invokes click's type-guessing logic. We don't want to do that since
        # it often incorrectly guesses that we want a composite type when we don't. [2]
        return str
    # Early out for homogenous containers (contains items of a single type)
    if len(args) == 1:
        return _clickify_arg(args[0])
    # Early out for homogenous tuples of indefinite length. E.g., `Tuple[int, ...]`.
    if len(args) == 2 and args[1] is Ellipsis:
        return _clickify_arg(args[0])
    # Last case is fixed-length containers (contains a fixed number of items of a
    # given type). E.g., `Tuple[str, int, int]`.
    return tuple(_clickify_args(args))


def _clickify_args(
    args: Tuple[type, ...],
) -> Iterable[SingleClickParamType]:
    return (_clickify_arg(arg) for arg in args)


def _clickify_arg(arg: type) -> SingleClickParamType:
    # When we don't know the type, we choose `str` (see [2])
    if arg is Any:
        return str
    # For containers and nested models, we use JSON
    if _is_container(arg) or issubclass(arg, BaseModel):
        return JSON_TYPE
    return arg


def _clickify_container_default(default: Any) -> Optional[Tuple[Any, ...]]:
    assert issubclass(type(default), Sequence)
    return tuple(v.json() if isinstance(v, BaseModel) else v for v in default)


def _type_name(type_: type) -> str:
    origin: Optional[type] = get_origin(type_)
    if origin is None:
        return type_.__name__
    return origin.__name__


def allow_if_specified(_: click.Context, param: click.Parameter, value: Any) -> Any:
    """Only allow options that the user explicitly specified."""
    if isinstance(param, _Option):
        return value if param.specified else None
    return value



from typing import Any, Dict


def deep_update(dest: Dict[Any, Any], other: Dict[Any, Any]) -> None:
    """Update `dest` with the key/value pairs from `other`.

    Returns `None`. Note that we modify `dest` in place.

    Unlike the built-in `dict.Update`, `deep_update` recurses into sub-dictionaries.
    This effectively "merges" `other` into `dest`.

    Note that we do not recurse into lists. We treat lists like any other
    non-`dict` type and simply override the existing entry in `dest` (if any).
    """
    for key, other_val in other.items():
        # Simple case: `key` is not in `dest`, so we simply add it.
        if key not in dest:
            dest[key] = other_val
            continue
        # Complex case: There is a conflict, so we must "merge" `dest_val`
        # and `other_val`.
        dest_val = dest[key]
        # Given two dicts, we can simply recurse.
        if isinstance(dest_val, dict) and isinstance(other_val, dict):
            deep_update(dest_val, other_val)
        # Any other type combination simply overrides the existing key in `dest`
        else:
            dest[key] = other_val



from __future__ import annotations

from pathlib import Path
from typing import Any, Callable, Dict, MutableMapping, Protocol

from pydantic import BaseSettings

from ....basic import deep_update as dict_deep_update

# Note that we disable D102 for `Protocol`s since it's redundant documentation.
# Similarly, we disable too-few-public-methods since it doens't make sense for
# `Protocol`s. Hopefully, both pydocstyle and pylint will special-case `Protocol`s
# soon enough.


class Loader(Protocol):  # pylint: disable=too-few-public-methods
    """Given raw file content, return a settings dict."""

    def __call__(self, __data: str) -> MutableMapping[str, Any]:  # noqa: D102
        ...


# We don't need that many public methods for this small utility class.
# Maybe we can replace it with a function that returns a function.
# I'm just not sure that this is a better approach.
class GlobSource:  # pylint: disable=too-few-public-methods
    """Find setting files in a directory with a glob pattern."""

    def __init__(
        self, directory: Path, pattern: str, loader: Loader, *, deep_update: bool = True
    ):
        self._dir = directory
        self._pattern = pattern
        self._loader = loader
        # We can't do better than `[Any, Any]` for now because of `dict.update`'s
        # lacking signature.
        self._update_func: Callable[[Any, Any], None]
        if deep_update:
            self._update_func = dict_deep_update
        else:
            self._update_func = dict.update

    def __call__(self, _: BaseSettings) -> Dict[str, Any]:
        """Return a dict with settings from the globbed files."""
        result: Dict[str, Any] = {}
        for path in sorted(self._dir.glob(self._pattern)):
            with path.open("r") as settings_file:
                # For now, we simply load everything into memory up front.
                # Settings files are usually small, so this shouldn't be
                # a problem in practice. In turn, it makes it easier to
                # support different loaders.
                data = settings_file.read()
            settings = dict(self._loader(data))
            self._update_func(result, settings)
        return result



try:
    from ._app import App
    from ._error import AppError
    from ._settings import Settings, autofill
except ImportError as exc:
    from .._extra import ExtraImportError

    raise ExtraImportError("app") from exc



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cyto',
 'cyto._extra',
 'cyto.app',
 'cyto.basic',
 'cyto.settings',
 'cyto.settings.sources',
 'cyto.settings.sources.cli',
 'cyto.settings.sources.glob']

package_data = \
{'': ['*']}

extras_require = \
{'all': ['anyio>=2.0.2,<3.0.0',
         'click>=7.1.2,<8.0.0',
         'pydantic>=1.8.1,<2.0.0',
         'toml>=0.10.2,<0.11.0'],
 'app': ['anyio>=2.0.2,<3.0.0', 'pydantic>=1.8.1,<2.0.0'],
 'settings': ['pydantic>=1.8.1,<2.0.0'],
 'settings.sources.cli': ['click>=7.1.2,<8.0.0'],
 'settings.sources.toml': ['toml>=0.10.2,<0.11.0']}

setup_kwargs = {
    'name': 'cyto',
    'version': '0.3.2',
    'description': "SBT Instruments' framework for Python-based applications",
    'long_description': '# Cyto 🦠\n\n## Idiomatic boilerplate and glue code for Python\n\nCyto lets you:\n * Create a [click](https://github.com/pallets/click) CLI from a [pydantic](https://github.com/kozlek/pydantic) model\n * Load [pydantic settings](https://pydantic-docs.helpmanual.io/usage/settings/) from [TOML](https://toml.io/en/) files [WIP]\n * [Daemonize](https://pagure.io/python-daemon/) your application [WIP]\n\nCyto is the bridge between an oppinionated selection of Python tech.\nCyto is everything you need to create a modern Python app.\nCyto is the glue and boilerplate code that you no longer have to write yourself.\n\n## Only pay for what you use\n\nSo you only want the click–pydantic integration but not the TOML loader?\nDon\'t worry, Cyto is split into so-called *extra*s so you can select just the extras that you want. In turn, Cyto only pulls in the dependencies for the selected extras. This is the "only pay for what you use" principle. In fact, Cyto has *zero* dependencies per default.\n\nHere is the list of extras:\n * `settings.sources.cli`: Create a [click](https://github.com/pallets/click) CLI from a [pydantic](https://github.com/kozlek/pydantic) model\n * `settings.sources.toml`: Load [pydantic settings](https://pydantic-docs.helpmanual.io/usage/settings/) from [TOML](https://toml.io/en/) files [WIP]\n * `app`: [WIP]\n\n## Installation\n\nInstall Cyto along with *all* extras:\n\n```\npip install cyto[all]\n```\n\nSimilar for poetry:\n\n```\npoetry add cyto[all]\n```\n\n### Choose specific extras\n\nIf you only want a specific extra, choose that when you install Cyto. E.g.:\n\n```\npip install cyto[settings.sources.cli]\n```\n\nSimilar for poetry:\n\n```\npoetry add cyto[settings.sources.cli]\n```\n\n## Development\n\n### Python Version\n\nDevelopment requires Python 3.8 or later. Test your python version with:\n```shell\npython3 --version\n```\n\nIf you have multiple python installations, you can replace `python3`\nwith a specific version (e.g., `python3.8`) in the steps below.\n\n### Getting Started\n\nDo the following:\n\n 1. Clone this repository\n    ```shell\n    git clone git@github.com:sbtinstruments/cyto.git\n    ```\n 2. Install poetry (for dependency management)\n    ```shell\n    curl -sSL https://raw.githubusercontent.com/python-poetry/poetry/master/get-poetry.py | python3\n    ```\n 3. Optional: Use a local poetry virtual environment\n    ```shell\n    poetry config --local virtualenvs.in-project true\n    ```\n    This integrates better with editors such as Visual Studio Code.\n 4. Create poetry\'s virtual environment and get all dependencies\n and all extra features.\n    ```shell\n    poetry install --extras all\n    ```\n 5. Optional: Run the QA basic tools (e.g., isort, black, pylint, mypy) automatically before each commit\n    ```shell\n    poetry run pre-commit install\n    ```\n\n### Quality Assurance (QA) Tools\n\n#### QA Basic Tools\n\n*All QA basic tools automatically run in Jenkins for each commit pushed\nto the remote repository. If you installed the `pre-commit` hooks,\nall QA basic tools automatically run before each commit too.*\n\nThe QA basic tools are:\n * `isort` (for import ordering)\n * `black` (for formatting)\n * `pylint` (for linting)\n * `mypy` (for type checking)\n\nYou can run the QA basic tools manually. This is useful if you\ndon\'t want to install the `pre-commit` hooks.\n\nRun the QA basic tools manually with:\n```shell\npoetry run task isort\npoetry run task black\npoetry run task pylint\npoetry run task mypy\n```\n\nRun all the basic QA tools manually with a single command:\n\n```shell\npoetry run task pre-commit\n```\n\nNote that this doesn\'t require you to install the `pre-commit` hooks.\n\n#### QA Test Tools\n\n*All of the tools below automatically run in Jenkins for each\ncommit pushed to the remote repository.*\n\nThe QA test tools are:\n * `tox` (for automation across Python versions)\n * `pytest` (the test framework itself)\n * `pytest-cov` (for test coverage percentage)\n\nRun the tests manually:\n\n 1. Install `tox`\n    ```shell\n    python3 -m pip install tox\n    ```\n 2. Start a tox run:\n    ```\n    tox\n    ```\n\nNote that `tox` invokes `pytest` in a set of virtual environments. Said\nvirtual environments have nothing to do with poetry\'s virtual environment. Poetry and tox runs in isolation of each other.\n\n### Visual Studio Code\n\n#### Settings\n\nWe have a default settings file that you can use via the following command:\n```shell\ncp .vscode/settings.json.default .vscode/settings.json\n```\nThis is optional.\n\n#### Python Interpreter\n\nHopefully, you used the local poetry virtual environment during\ninstallation (the `poetry config --local virtualenvs.in-project true` part). This way, Visual Studio Code automatically finds the\nPython interpreter within poetry\'s virtual environment.\n\nAlternatively, you can point Visual Studio Code to poetry\'s\nglobal virtual environments folder. Add the following entry\nto your `./vscode/settings.json` file:\n```json\n{ "python.venvPath": "~/.cache/pypoetry/virtualenvs" }\n```\n\nThen, you look for the poetry\'s currently active virtual environment:\n```shell\npoetry env list\n```\n\nLastly, you use the Visual Studio Code command\n`Python: Select Interpreter` and choose the interpreter inside\npoetry\'s currently active virtual environment.',
    'author': 'Frederik Peter Aalund',
    'author_email': 'fpa@sbtinstruments.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/sbtinstruments/cyto',
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
